"""Interface for querying and controlling thermostats."""

from decimal import Decimal
from enum import IntEnum
from typing import Any, Dict, List

from .base import Interface


class ThermostatInterface(Interface):
    """Interface for querying and controlling thermostats."""

    class OperationMode(IntEnum):
        """The operation mode of the thermostat."""

        OFF = 0
        COOL = 1
        HEAT = 2
        AUTO = 3
        UNKNOWN = 4

    class FanMode(IntEnum):
        """The fan mode of the thermostat."""

        OFF = 0
        ON = 1
        UNKNOWN = 2

    class DayMode(IntEnum):
        """The day mode of the thermostat."""

        DAY = 0
        NIGHT = 1
        UNKNOWN = 2
        STANDBY = 3

    class HoldMode(IntEnum):
        """The hold mode of the thermostat."""

        NORMAL = 0
        HOLD = 1
        UNKNOWN = 2

    class Status(IntEnum):
        """The status of the thermostat."""

        OFF = 0
        COOLING = 1
        HEATING = 2
        OFFLINE = 3

    async def get_indoor_temperature(self, vid: int) -> Decimal:
        """Get the current indoor temperature.

        Args:
            vid: The Vantage ID of the thermostat.
        """
        # INVOKE <id> Thermostat.GetIndoorTemperature
        # -> R:INVOKE <id> <temp> Thermostat.GetIndoorTemperature
        response = await self.invoke(vid, "Thermostat.GetIndoorTemperature")
        temp = Decimal(response.args[1])

        return temp

    async def get_outdoor_temperature(self, vid: int) -> Decimal:
        """Get the current outdoor temperature.

        Args:
            vid: The Vantage ID of the thermostat.
        """
        # INVOKE <id> Thermostat.GetOutdoorTemperature
        # -> R:INVOKE <id> <temp> Thermostat.GetOutdoorTemperature
        response = await self.invoke(vid, "Thermostat.GetOutdoorTemperature")
        temp = Decimal(response.args[1])

        return temp

    async def get_heat_set_point(self, vid: int) -> Decimal:
        """Get the current heat set point.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The heat set point of the thermostat, in degrees Celsius.
        """
        # INVOKE <id> Thermostat.GetHeatSetPoint
        # -> R:INVOKE <id> <temp> Thermostat.GetHeatSetPoint
        response = await self.invoke(vid, "Thermostat.GetHeatSetPoint")
        temp = Decimal(response.args[1])

        return temp

    async def set_heat_set_point(self, vid: int, temp: float) -> None:
        """Set the current heat set point.

        Args:
            vid: The Vantage ID of the thermostat.
            temp: The heat set point to set, in degrees Celsius.
        """
        # INVOKE <id> Thermostat.SetHeatSetPoint <temp>
        # -> R:INVOKE <id> Thermostat.SetHeatSetPoint <temp>
        await self.invoke(vid, "Thermostat.SetHeatSetPoint", temp)

    async def get_cool_set_point(self, vid: int) -> Decimal:
        """Get the current cool set point.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The cool set point of the thermostat, in degrees Celsius.
        """
        # INVOKE <id> Thermostat.GetCoolSetPoint
        # -> R:INVOKE <id> <temp> Thermostat.GetCoolSetPoint
        response = await self.invoke(vid, "Thermostat.GetCoolSetPoint")
        temp = Decimal(response.args[1])

        return temp

    async def set_cool_set_point(self, vid: int, temp: float) -> None:
        """Set the current cool set point.

        Args:
            vid: The Vantage ID of the thermostat.
            temp: The cool set point to set, in degrees Celsius.
        """
        # INVOKE <id> Thermostat.SetCoolSetPoint <temp>
        # -> R:INVOKE <id> Thermostat.SetCoolSetPoint <temp>
        await self.invoke(vid, "Thermostat.SetCoolSetPoint", temp)

    async def get_operation_mode(self, vid: int) -> OperationMode:
        """Get the current operation mode.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The operation mode of the thermostat.
        """
        # INVOKE <id> Thermostat.GetOperationMode
        # -> R:INVOKE <id> <mode (Off|Cool|Heat|Auto|Unknown)> Thermostat.GetOperationMode
        response = await self.invoke(vid, "Thermostat.GetOperationMode")
        if response.args[1].isdigit():
            mode = self.OperationMode(int(response.args[1]))
        else:
            mode = self.OperationMode[response.args[1].upper()]

        return mode

    async def set_operation_mode(self, vid: int, mode: int) -> None:
        """Set the current operation mode.

        Args:
            vid: The Vantage ID of the thermostat.
            mode: The operation mode to set.
        """
        # INVOKE <id> Thermostat.SetOperationMode <mode>
        # -> R:INVOKE <id> Thermostat.SetOperationMode <mode>
        await self.invoke(vid, "Thermostat.SetOperationMode", mode)

    async def get_fan_mode(self, vid: int) -> FanMode:
        """Get the current fan mode.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The fan mode of the thermostat.
        """
        # INVOKE <id> Thermostat.GetFanMode
        # -> R:INVOKE <id> <mode (Off|On|Unknown)> Thermostat.GetFanMode
        response = await self.invoke(vid, "Thermostat.GetFanMode")
        if response.args[1].isdigit():
            mode = self.FanMode(int(response.args[1]))
        else:
            mode = self.FanMode[response.args[1].upper()]

        return mode

    async def set_fan_mode(self, vid: int, mode: int) -> None:
        """Set the current fan mode.

        Args:
            vid: The Vantage ID of the thermostat.
            mode: The fan mode to set.
        """
        # INVOKE <id> Thermostat.SetFanMode <mode>
        # -> R:INVOKE <id> Thermostat.SetFanMode <mode>
        await self.invoke(vid, "Thermostat.SetFanMode", mode)

    async def get_day_mode(self, vid: int) -> DayMode:
        """Get the current day mode.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The day mode of the thermostat.
        """
        # INVOKE <id> Thermostat.GetDayMode
        # -> R:INVOKE <id> <mode (Day|Night|Unknown|Standby)> Thermostat.GetDayMode
        response = await self.invoke(vid, "Thermostat.GetDayMode")
        if response.args[1].isdigit():
            mode = self.DayMode(int(response.args[1]))
        else:
            mode = self.DayMode[response.args[1].upper()]

        return mode

    async def set_day_mode(self, vid: int, mode: int) -> None:
        """Set the current day mode.

        Args:
            vid: The Vantage ID of the thermostat.
            mode: The day mode to set.
        """
        # INVOKE <id> Thermostat.SetDayMode <mode>
        # -> R:INVOKE <id> Thermostat.SetDayMode <mode>
        await self.invoke(vid, "Thermostat.SetDayMode", mode)

    async def get_hold_mode(self, vid: int) -> HoldMode:
        """Get the current hold mode.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The hold mode of the thermostat.
        """
        # INVOKE <id> Thermostat.GetHoldMode
        # -> R:INVOKE <id> <mode> Thermostat.GetHoldMode
        response = await self.invoke(vid, "Thermostat.GetHoldMode")
        if response.args[1].isdigit():
            mode = self.HoldMode(int(response.args[1]))
        else:
            mode = self.HoldMode[response.args[1].upper()]

        return mode

    async def set_hold_mode(self, vid: int, mode: int) -> None:
        """Set the current hold mode.

        Args:
            vid: The Vantage ID of the thermostat.
            mode: The hold mode to set.
        """
        # INVOKE <id> Thermostat.SetHoldMode <mode>
        # -> R:INVOKE <id> Thermostat.SetHoldMode <mode>
        await self.invoke(vid, "Thermostat.SetHoldMode", mode)

    async def get_status(self, vid: int) -> Status:
        """Get the current status.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The status of the thermostat.
        """
        # INVOKE <id> Thermostat.GetStatus
        # -> R:INVOKE <id> <status> Thermostat.GetStatus
        response = await self.invoke(vid, "Thermostat.GetStatus")
        if response.args[1].isdigit():
            status = self.Status(int(response.args[1]))
        else:
            status = self.Status[response.args[1].upper()]

        return status

    async def get_auto_set_point(self, vid: int) -> Decimal:
        """Get the current auto set point.

        Args:
            vid: The Vantage ID of the thermostat.

        Returns:
            The auto set point of the thermostat, in degrees Celsius.
        """
        # INVOKE <id> Thermostat.GetAutoSetPoint
        # -> R:INVOKE <id> <temp> Thermostat.GetAutoSetPoint
        response = await self.invoke(vid, "Thermostat.GetAutoSetPoint")
        temp = Decimal(response.args[1])

        return temp

    async def set_auto_set_point(self, vid: int, temp: float) -> None:
        """Set the current auto set point.

        Args:
            vid: The Vantage ID of the thermostat.
            temp: The auto set point to set, in degrees Celsius.
        """
        # INVOKE <id> Thermostat.SetAutoSetPoint <temp>
        # -> R:INVOKE <id> Thermostat.SetAutoSetPoint <temp>
        await self.invoke(vid, "Thermostat.SetAutoSetPoint", temp)

    @classmethod
    def parse_status(cls, _vid: int, method: str, args: List[str]) -> Dict[str, Any]:
        """Parse a Thermostat status event."""
        if method == "Thermostat.GetIndoorTemperature":
            # ADDSTATUS <id>
            # -> S:STATUS <id> Thermostat.GetIndoorTemperature <temp>
            # ELLOG STATUSEX ON
            # -> EL: <id> Thermostat.GetIndoorTemperature <temp>
            return {
                "indoor_temperature": Decimal(args[0]),
            }

        if method == "Thermostat.GetOutdoorTemperature":
            # ADDSTATUS <id>
            # -> S:STATUS <id> Thermostat.GetOutdoorTemperature <temp>
            # ELLOG STATUSEX ON
            # -> EL: <id> Thermostat.GetOutdoorTemperature <temp>
            return {
                "outdoor_temperature": Decimal(args[0]),
            }

        if method == "Thermostat.GetHeatSetPoint":
            # ADDSTATUS <id>
            # -> S:STATUS <id> Thermostat.GetHeatSetPoint <temp>
            # ELLOG STATUSEX ON
            # -> EL: <id> Thermostat.GetHeatSetPoint <temp>
            return {
                "heat_set_point": Decimal(args[0]),
            }

        if method == "Thermostat.GetCoolSetPoint":
            # ADDSTATUS <id>
            # -> S:STATUS <id> Thermostat.GetCoolSetPoint <temp>
            # ELLOG STATUSEX ON
            # -> EL: <id> Thermostat.GetCoolSetPoint <temp>
            return {
                "cool_set_point": Decimal(args[0]),
            }

        if method == "Thermostat.GetOperationMode":
            # ADDSTATUS <id>
            # -> S:STATUS <id> Thermostat.GetOperationMode <mode>
            # ELLOG STATUSEX ON
            # -> EL: <id> Thermostat.GetOperationMode <mode>
            return {
                "operation_mode": cls.OperationMode(int(args[0])),
            }

        if method == "Thermostat.GetFanMode":
            # ADDSTATUS <id>
            # -> S:STATUS <id> Thermostat.GetFanMode <mode>
            # ELLOG STATUSEX ON
            # -> EL: <id> Thermostat.GetFanMode <mode>
            return {
                "fan_mode": cls.FanMode(int(args[0])),
            }

        if method == "Thermostat.GetDayMode":
            # ADDSTATUS <id>
            # -> S:STATUS <id> Thermostat.GetDayMode <mode>
            # ELLOG STATUSEX ON
            # -> EL: <id> Thermostat.GetDayMode <mode>
            return {
                "day_mode": cls.DayMode(int(args[0])),
            }

        if method == "Thermostat.GetHoldMode":
            # ADDSTATUS <id>
            # -> S:STATUS <id> Thermostat.GetHoldMode <mode>
            # ELLOG STATUSEX ON
            # -> EL: <id> Thermostat.GetHoldMode <mode>
            return {
                "hold_mode": cls.HoldMode(int(args[0])),
            }

        if method == "Thermostat.GetStatus":
            # ADDSTATUS <id>
            # -> S:STATUS <id> Thermostat.GetStatus <status>
            # ELLOG STATUSEX ON
            # -> EL: <id> Thermostat.GetStatus <status>
            return {
                "status": cls.Status(int(args[0])),
            }

        if method == "Thermostat.GetAutoSetPoint":
            # ADDSTATUS <id>
            # -> S:STATUS <id> Thermostat.GetAutoSetPoint <temp>
            # ELLOG STATUSEX ON
            # -> EL: <id> Thermostat.GetAutoSetPoint <temp>
            return {
                "auto_set_point": Decimal(args[0]),
            }

        return {}



"""Controller holding and managing Vantage light sensors."""

from decimal import Decimal
from typing import Sequence

from typing_extensions import override

from aiovantage.command_client.object_interfaces import LightSensorInterface
from aiovantage.models import LightSensor

from .base import BaseController, State


class LightSensorsController(BaseController[LightSensor], LightSensorInterface):
    """Controller holding and managing Vantage light sensors."""

    vantage_types = ("LightSensor",)
    """The Vantage object types that this controller will fetch."""

    status_types = ("LIGHT",)
    """Which Vantage 'STATUS' types this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of a light sensor."""
        return {
            "level": await LightSensorInterface.get_level(self, vid),
        }

    @override
    def parse_object_update(self, _vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for a light sensor."""
        if status != "LIGHT":
            return None

        # STATUS LIGHT
        # -> S:LIGHT <id> <level>
        return {
            "level": Decimal(args[0]),
        }



"""Controller holding and managing Vantage tasks."""

from typing import Sequence

from typing_extensions import override

from aiovantage.command_client.object_interfaces import TaskInterface
from aiovantage.models import Task

from .base import BaseController, State


class TasksController(BaseController[Task], TaskInterface):
    """Controller holding and managing Vantage tasks."""

    vantage_types = ("Task",)
    """The Vantage object types that this controller will fetch."""

    status_types = ("TASK",)
    """Which Vantage 'STATUS' types this controller handles, if any."""

    object_status = True
    """Should this controller subscribe to "object status" events."""

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of a task."""
        return {
            "is_running": await TaskInterface.is_running(self, vid),
            "state": await TaskInterface.get_state(self, vid),
        }

    @override
    def parse_object_update(self, _vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for a task."""
        if status == "TASK":
            # STATUS TASK
            # -> S:TASK <id> <state>
            return {
                "state": int(args[0]),
            }

        if status == "Task.IsRunning":
            return {
                "is_running": TaskInterface.parse_is_running_status(args),
            }

        return None



"""Light sensor object."""

from dataclasses import dataclass, field
from decimal import Decimal
from typing import Optional

from .sensor import Sensor
from .types import Parent


@dataclass
class LightSensor(Sensor):
    """Light sensor object."""

    parent: Parent = field(
        metadata={
            "name": "Parent",
        }
    )

    level: Optional[Decimal] = field(
        default=None,
        metadata={
            "type": "Ignore",
        },
    )



"""Controller holding and managing Vantage controllers."""

from typing_extensions import override

from aiovantage.command_client.object_interfaces import IntrospectionInterface
from aiovantage.models import Master

from .base import BaseController, State


class MastersController(BaseController[Master], IntrospectionInterface):
    """Controller holding and managing Vantage controllers."""

    vantage_types = ("Master",)
    """The Vantage object types that this controller will fetch."""

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of a Vantage controller."""
        return {
            "firmware_version": await IntrospectionInterface.get_firmware_version(
                self, vid, IntrospectionInterface.Firmware.APPLICATION
            ),
        }



"""IIntrospection.GetSysInfo method definition."""

from dataclasses import dataclass, field
from typing import ClassVar, Optional


@dataclass
class SysInfo:
    """Partial SysInfo definition."""

    master_number: int = field(metadata={"name": "MasterNumber"})
    serial_number: int = field(metadata={"name": "SerialNumber"})


@dataclass
class GetSysInfo:
    """IIntrospection.GetSysInfo method definition."""

    interface: ClassVar[str] = "IIntrospection"
    call = None
    return_value: Optional["GetSysInfo.Return"] = field(
        default=None,
        metadata={
            "name": "return",
        },
    )

    @dataclass
    class Return:
        """IIntrospection.GetSysInfo method return value."""

        sys_info: SysInfo = field(metadata={"name": "SysInfo"})



"""IIntrospection.GetTypes method definition."""

from dataclasses import dataclass, field
from typing import ClassVar, List, Optional


@dataclass
class Type:
    """Object type definition."""

    name: str = field(metadata={"name": "Name"})
    version: str = field(metadata={"name": "Version"})


@dataclass
class GetTypes:
    """IIntrospection.GetTypes method definition."""

    interface: ClassVar[str] = "IIntrospection"
    call = None
    return_value: Optional[List[Type]] = field(
        default=None,
        metadata={
            "name": "Type",
            "wrapper": "return",
        },
    )



"""IIntrospection.GetInterfaces method definition."""

from dataclasses import dataclass, field
from typing import ClassVar, List, Optional


@dataclass
class Interface:
    """Object interface definition."""

    name: str = field(metadata={"name": "Name"})
    version: str = field(metadata={"name": "Version"})
    id: int = field(metadata={"name": "IID"})


@dataclass
class GetInterfaces:
    """IIntrospection.GetInterfaces method definition."""

    interface: ClassVar[str] = "IIntrospection"
    call = None
    return_value: Optional[List[Interface]] = field(
        default=None,
        metadata={
            "name": "Interface",
            "wrapper": "return",
        },
    )



"""Controller holding and managing Vantage omni sensors."""

from decimal import Decimal
from typing import Sequence, Union

from typing_extensions import override

from aiovantage.config_client.models.omni_sensor import ConversionType, OmniSensor

from .base import BaseController, State


class OmniSensorsController(BaseController[OmniSensor]):
    """Controller holding and managing Vantage omni sensors.

    Omni sensors are generic sensors objects which specify which methods to use
    when getting or setting data in their object definition, as well as the
    type of data and a conversion formula.
    """

    vantage_types = ("OmniSensor",)
    """The Vantage object types that this controller will fetch."""

    object_status = True
    """Should this controller subscribe to "object status" events."""

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of an omni sensor."""
        return {
            "level": await self.get_level(vid),
        }

    @override
    def parse_object_update(self, vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for an omni sensor."""
        omni_sensor = self[vid]
        if status != omni_sensor.get.method:
            return None

        return {
            "level": self.parse_get_level_status(omni_sensor, args),
        }

    async def get_level(self, vid: int, cached: bool = False) -> Union[int, Decimal]:
        """Get the level of an OmniSensor.

        Args:
            vid: The ID of the sensor.
            cached: Whether to use the cached value or fetch a new one.

        Returns:
            The level of the sensor.
        """
        # Figure out which get method to use, hardware or software (cached)
        omni_sensor = self[vid]
        method = omni_sensor.get.method if cached else omni_sensor.get.method_hw

        # INVOKE <id> <method>
        # -> R:INVOKE <id> <value> <method>
        response = await self.command_client.command("INVOKE", vid, method)

        # Convert the level to the correct type
        if omni_sensor.get.formula.return_type == ConversionType.FIXED:
            level = Decimal(response.args[1])
            if omni_sensor.get.formula.level_type == ConversionType.INT:
                return int(level)

            return level

        if omni_sensor.get.formula.return_type == ConversionType.INT:
            level = Decimal(response.args[1]) / 1000
            if omni_sensor.get.formula.level_type == ConversionType.INT:
                return int(level)

            return level

        raise ValueError(f"Unknown return type {omni_sensor.get.formula.return_type}")

    @classmethod
    def parse_get_level_status(
        cls, omni_sensor: OmniSensor, args: Sequence[str]
    ) -> Union[int, Decimal]:
        """Parse an OmniSensor 'GetLevel' event, eg. 'PowerSensor.GetPower'."""
        # ELLOG STATUS ON
        # -> EL: <id> <method> <value>
        # STATUS ADD <id>
        # -> S:STATUS <id> <method> <value>
        level = Decimal(args[0]) / 1000
        if omni_sensor.get.formula.level_type == ConversionType.INT:
            return int(level)

        return level



"""Controller holding and managing thermostats."""

from typing import Sequence

from typing_extensions import override

from aiovantage.command_client.object_interfaces import ThermostatInterface
from aiovantage.models import Temperature, Thermostat
from aiovantage.query import QuerySet

from .base import BaseController, State


class ThermostatsController(BaseController[Thermostat], ThermostatInterface):
    """Controller holding and managing thermostats."""

    vantage_types = ("Thermostat",)
    """The Vantage object types that this controller will fetch."""

    status_types = ("THERMFAN", "THERMOP", "THERMDAY")
    """Which Vantage 'STATUS' types this controller handles, if any."""

    @override
    async def fetch_object_state(self, vid: int) -> State:
        """Fetch the state properties of a thermostat."""
        return {
            "day_mode": await ThermostatInterface.get_day_mode(self, vid),
            "fan_mode": await ThermostatInterface.get_fan_mode(self, vid),
            "operation_mode": await ThermostatInterface.get_operation_mode(self, vid),
        }

    @override
    def parse_object_update(self, _vid: int, status: str, args: Sequence[str]) -> State:
        """Handle state changes for a thermostat."""
        if status == "THERMFAN":
            # STATUS THERMFAN
            # -> S:THERMFAN <id> <fan_mode (ON/AUTO)>
            if args[0] == "ON":
                fan_mode = Thermostat.FanMode.ON
            elif args[0] == "AUTO":
                fan_mode = Thermostat.FanMode.OFF

            return {
                "fan_mode": fan_mode,
            }

        if status == "THERMOP":
            # STATUS THERMOP
            # -> S:THERMOP <id> <operation_mode (OFF/COOL/HEAT/AUTO)>
            if args[0] == "OFF":
                op_mode = Thermostat.OperationMode.OFF
            elif args[0] == "COOL":
                op_mode = Thermostat.OperationMode.COOL
            elif args[0] == "HEAT":
                op_mode = Thermostat.OperationMode.HEAT
            elif args[0] == "AUTO":
                op_mode = Thermostat.OperationMode.AUTO

            return {
                "operation_mode": op_mode,
            }

        if status == "THERMDAY":
            # STATUS THERMDAY
            # -> S:THERMDAY <id> <day_mode (DAY/NIGHT)>
            if args[0] == "DAY":
                day_mode = Thermostat.DayMode.DAY
            elif args[0] == "NIGHT":
                day_mode = Thermostat.DayMode.NIGHT

            return {
                "day_mode": day_mode,
            }

        return None

    def sensors(self, vid: int) -> QuerySet[Temperature]:
        """Return all sensors associated with this thermostat."""
        return self._vantage.temperature_sensors.filter(
            lambda obj: obj.parent.id == vid
        )

    def indoor_sensor(self, vid: int) -> QuerySet[Temperature]:
        """Return a queryset to fetch the indoor temperature sensor for this thermostat."""
        return self.sensors(vid).filter(lambda obj: obj.parent.position == 1)

    def outdoor_sensor(self, vid: int) -> QuerySet[Temperature]:
        """Return a queryset to fetch the outdoor temperature sensor for this thermostat."""
        return self.sensors(vid).filter(lambda obj: obj.parent.position == 2)

    def cool_setpoint(self, vid: int) -> QuerySet[Temperature]:
        """Return a queryset to fetch the cool setpoint sensor for this thermostat."""
        return self.sensors(vid).filter(
            lambda obj: obj.setpoint == Temperature.Setpoint.COOL
            or obj.parent.position == 3
        )

    def heat_setpoint(self, vid: int) -> QuerySet[Temperature]:
        """Return a queryset to fetch the heat setpoint sensor for this thermostat."""
        return self.sensors(vid).filter(
            lambda obj: obj.setpoint == Temperature.Setpoint.HEAT
            or obj.parent.position == 4
        )
